from ..utils import ViewUtil
import hashlib,os,time,types,inspect
from .TemplateCompiler import TemplateCompiler
from .Filters import Filters

"""
 * 模板基类
 *<B>说明：</B>
 *<pre>
 *  使用
 *</pre>
"""
class BaseTemplate:

    # 函数过滤器对象
    filters = None;

    # 上下文对象
    context = None;

    # 模板代码缓存字典
    caches = {}


    def __init__(self, attrs={}):

        # 模板文件扩展名
        self.suffix = 'html'

        # 模板文件根路径
        self.tplPath = '';

        # 是否开启模板缓存,开启后,会自动缓存代码,加快模板解析的速度
        self.onCache = True;

        # 模板缓存文件目录,模板编译的python代码缓存在文件中,缓存文件存储此目录中
        self.cachePath = '';

        # 资源地址,比如js,css,img,默认提供static(静态资源路径),res(外部资源,比如上传的文件) 字典key
        self.urls = {}

        # 模板缓存有效期,单位秒,0 表示无有效期
        self.timeout = 3;

        # 表达式起始符
        self.expStart = '{{';
        # 表达式结束符
        self.expEnd = '}}'

        # 结束表达式的结束符比如/ 则完整表达式为{/for} 或end,{endfor}
        self.expEndSign = '/'

        # 是否启用标签规则,开启后,匹配表达式 <import name="eduhome.name.yong" />
        self.onTag = True

        # 标签起始符
        self.tagStart = '<'
        # 标签结束符
        self.tagEnd = '>'

        # 结束标签结束符号
        self.tagEndSign = '/'

        # 系统标签,默认自动加载sys 表达式,默认标签,书写时无需写入前缀,比如<css href="xxx" />
        self.sysTags = [];

        # 自定义标签,书写时必须写入前缀,比如 标签名称html,则css 标签的书写规则为: <html:css href="xxx" />
        self.customTags = [];

        # 当前模板数据
        self.data = {};

        # 默认表达式标签对象列表
        self.defaultExpressions = []

        # 系统标签对象
        self.defaultSysTag = None;

        self.contextData = {}; # 当前模板上下文数据

        if attrs:
            ViewUtil.setAttrs(self, attrs)

        self.__initView()

    def __initView(self):

        if self.__class__.filters.template is None:
            self.__class__.filters.setTemplate(self)

        staticUrl = self.urls.get('static', None)
        if staticUrl is None:
            self.urls['static'] = '';

        resUrl = self.urls.get('res', None)
        if resUrl is None:
            self.urls['res'] = '';

        return ;

    # 验证过滤器是否存在
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def hasfilters(self, func):

        return hasattr(self.__class__.filters, func)

    # 获取模板预加载的标签库表达式对象
    # <B> 说明： </B>
    # <pre>
    # 不包括模板文件导入的标签
    # </pre>
    def getTaglibExpression(self)->'list':

        if self.defaultExpressions:
            return self.defaultExpressions;

        expList = [];
        from ..tag.Tag import Tag

        defaultSysTag = Tag.make('sys', self)
        self.defaultSysTag = defaultSysTag;

        expList = defaultSysTag.getTagExpression();

        # 系统标签
        for tagName in self.sysTags:
            expList = expList + self.buildTaglibExpression(tagName,True)

        # 自定义标签
        for tagName in  self.customTags:
            expList = expList + self.buildTaglibExpression(tagName)


        self.defaultExpressions = expList

        return self.defaultExpressions

    # 获取指定标签库的所有表达式对象
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def buildTaglibExpression(self,tagName,sysTag = False)->'list':

        from ..tag.Tag import Tag
        tagObj = Tag.make(tagName, self)

        tagAlias = ''
        if sysTag is False:
            tagAlias = tagObj.getAlias();
            if not tagAlias:
                tagAlias = tagName

        return tagObj.getTagExpression(tagAlias)

    # 注册模板过滤器
    # <B> 说明： </B>
    # <pre>
    # 对外唯一入口
    # </pre>
    @classmethod
    def registerFilter(cls,funcName,funcMeta):

        if cls.filters is None:
            cls.filters = Filters()

        signature = inspect.signature(funcMeta)
        firstArg = list(signature.parameters.keys())[0];
        if firstArg == 'self':
            setattr(cls.filters, funcName,  types.MethodType(funcMeta, cls.filters))
        else:
            setattr(cls.filters, funcName,funcMeta)

    # 注册模板上下文
    # <B> 说明： </B>
    # <pre>
    # 对外唯一入口
    # </pre>
    @classmethod
    def registerContext(cls, funcName, funcMeta):

        if cls.context is None:
            from .TemplateContext import TemplateContext
            cls.context = TemplateContext()

        cls.context.register(funcName,funcMeta)


    # 注入模板参数
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def assign(self, name, value):

        self.data[name] = value

        return ;

    # 加载模板并返回执行模板结果
    # <B> 说明： </B>
    # <pre>
    # 对外唯一入口
    # </pre>
    def fetch(self, tpl, data = {}):

        self.data.update(data)

        return self.loadTemplate(tpl)

    # 加载模板
    # <B> 说明： </B>
    # <pre>
    # 内部调用,目前用于小物件
    # </pre>
    def render(self,tpl,data = {}):

        # 扩展名处理
        tpl = self._buildTempExt(tpl)
        viewTplCode = self.compile(tpl)
        result = ''
        mydata = self._mergeContext(data);
        from .MyTemplate import MyTemplate
        mydata['_result'] = result
        mydata['_this'] = self
        mydata['_MyTemplate'] = MyTemplate
        mydata['FI'] = self.__class__.filters

        exec(viewTplCode, mydata)

        return mydata['_result']

    # 加载,执行模板模板
    # <B> 说明： </B>
    # <pre>
    # 内部调动,目前用于include,layout 标签,
    # </pre>
    def loadTemplate(self, tpl):
        # 扩展名处理
        tpl = self._buildTempExt(tpl)
        viewTplCode = self.compile(tpl)
        result = ''

        mydata = self._mergeContext(self.data);
        from .MyTemplate import MyTemplate
        mydata['_result'] = result
        mydata['_this'] = self
        mydata['_MyTemplate'] = MyTemplate
        mydata['FI'] = self.__class__.filters
        exec(viewTplCode,mydata)

        return mydata['_result']

    # 检测模板缓存文件有效
    # <B> 说明： </B>
    # <pre>
    # True: 表示模板缓存有效
    # False:表示模板缓存失效
    # </pre>
    def checkTemplateCacheFile(self, templateFile = ''):

        if self.onCache is not True:
            return False

        tplFilePath = self._buildTemplateFilePath(templateFile)
        tplCacheFilePath = self._buildTemplateCacheFilePath(templateFile)

        if not os.path.exists(tplFilePath):
            raise RuntimeError('template {0} file is not exist'.format(templateFile))

        if not os.path.exists(tplCacheFilePath):
            return False

        tplFilePathStat = os.stat(tplFilePath)
        tplCacheFilePathStat = os.stat(tplCacheFilePath)
        tplFilePathUptime = tplFilePathStat.st_mtime

        # 校验修改事件
        if tplFilePathUptime > tplCacheFilePathStat.st_mtime:
            return False
        # 校验有效期
        nowTime = int(time.time())

        if self.timeout > 0 and nowTime > (int(tplCacheFilePathStat.st_mtime) + self.timeout):
            return False

        return True

    # 检测代码缓存是否过期
    # <B> 说明： </B>
    # <pre>
    # True: 表示模板缓存有效
    # False:表示模板缓存失效
    # </pre>
    def checkTemplateCacheCode(self, templateFile):

        tplFilePath = self._buildTemplateFilePath(templateFile)
        if not os.path.exists(tplFilePath):
            raise RuntimeError('template {0} file is not exist'.format(templateFile))

        tplFilePathUptime = os.stat(tplFilePath).st_mtime
        cacheCode = self._getTemplateContentFromCache(templateFile);

        if not cacheCode:
            return False

        cacheCodeTime = cacheCode['time']

        if tplFilePathUptime > cacheCodeTime:
            return False

        nowTime = int(time.time())

        if self.timeout > 0 and nowTime > (cacheCodeTime + self.timeout):
            return False

        return True;

    # 合并上下文数据
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def _mergeContext(self,data):

        if self.__class__.context and not self.contextData:
            self.contextData = self.__class__.context.getContext()

        mydata = {}
        mydata.update(self.contextData);
        mydata.update(data);

        return mydata;

    # 构建模板扩展名
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def _buildTempExt(self,tpl):

        tplExt = ViewUtil.getFileExt(tpl);
        if not tplExt:
            tpl = '{0}.{1}'.format(tpl,self.suffix)

        return tpl

    # 编译模板
    # <B> 说明： </B>
    # <pre>
    # 单纯的编译模板文件,不执行python 代码
    # </pre>
    def compile(self, templateFile):

        if self.onCache and (self.cachePath and self.checkTemplateCacheFile(templateFile)) or (not self.cachePath and self.checkTemplateCacheCode(templateFile)):
            cacheCode = self._getTemplateContentFromCache(templateFile)
            if not cacheCode:
                # 缓存有效
                tplCacheFilePath = self._buildTemplateCacheFilePath(templateFile)
                with open(tplCacheFilePath, 'r') as f:
                    viewTplCode = f.read()

                self._writeTemplateContentToCache(templateFile,viewTplCode)

            else:
                viewTplCode = cacheCode['code']

        else:
            templateContent = self.__compileTemplate(templateFile);
            viewTplCode = self._buildFullTemplateCode(templateContent, templateFile)

            # 编译代码写入内存缓存
            if self.cachePath:
                tplCacheFilePath = self._buildTemplateCacheFilePath(templateFile)
                with open(tplCacheFilePath, 'w') as f:
                    f.write(viewTplCode)

            # 编译代码写入内存缓存
            self._writeTemplateContentToCache(templateFile,viewTplCode)

        return viewTplCode

    def __compileTemplate(self,templateFile)->'str':

        tplCompilerContent = self._readTemplateContent(templateFile);
        templateCompiler = self.makeTemplateCompiler()
        templateContent = templateCompiler.compiler(tplCompilerContent)

        return templateContent

    # 读取模板内容
    def _readTemplateContent(self,templateFile)->'str':

        tplFilePath = self._buildTemplateFilePath(templateFile)
        tplCompilerContent = ''
        with open(tplFilePath, 'r') as f:
            tplCompilerContent = f.read()

        lines = tplCompilerContent.splitlines()
        tplCompilerContent = r'\n'.join(lines)

        return tplCompilerContent

    # 创建模板编译对象
    # <B> 说明： </B>
    # <pre>
    # 用于编译模板
    # </pre>
    def makeTemplateCompiler(self) -> 'TemplateCompiler':

        return TemplateCompiler(self)

    # 从缓存获取模板内容
    # <B> 说明： </B>
    # <pre>
    #
    # </pre>
    def _getTemplateContentFromCache(self,templateFile):

        cacheKey = self._buildCacheKey(templateFile)

        return self.__class__.caches.get(cacheKey,None)

    def _writeTemplateContentToCache(self,templateFile,tplCompilerContent):

        cacheKey = self._buildCacheKey(templateFile)
        self.__class__.caches[cacheKey] = {"time":int(time.time()),"code":tplCompilerContent}

        return ;

    def _buildCacheKey(self,templateFile):

        tplFilePath = self._buildTemplateFilePath(templateFile).encode("utf8");
        tplFileHash = hashlib.md5(tplFilePath)

        return tplFileHash.hexdigest()



    # 构建模板文件绝对路径
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def _buildTemplateFilePath(self, templateFile):

        tplFilePath = self.tplPath + templateFile

        return tplFilePath

    # 构建模板缓存文件绝对路径
    # <B> 说明： </B>
    # <pre>
    # True: 表示模板缓存有效
    # False:表示模板缓存失效
    # </pre>
    def _buildTemplateCacheFilePath(self, templateFile):

        tempFileList = templateFile.split('.')[0].split('/');
        tempFile = '_'.join(tempFileList);
        tplFile = self._buildTemplateFilePath(templateFile).encode("utf8");
        tplFileHash = hashlib.md5(tplFile)

        return '{0}{1}_{2}.py'.format(self.cachePath, tplFileHash.hexdigest(),tempFile)

    def _buildTemplateClazz(self, templateFile):

        tplFile = self._buildTemplateFilePath(templateFile).encode("utf8");
        tplFileHash = hashlib.md5(tplFile)
        templateClazz = 'view{0}'.format(tplFileHash.hexdigest());

        return templateClazz

    def _buildFullTemplateCode(self, compileContent, templateFile):

        templateClazz = self._buildTemplateClazz(templateFile)
        with open(os.path.dirname(os.path.realpath(__file__)) + '/view.html', 'r') as f:
            viewTplCode = f.read()
            viewTplCode = viewTplCode.format(clazz=templateClazz, body=compileContent)

        return viewTplCode


from .BaseNode import BaseNode

"""
 * 方法节点
 *<B>说明：</B>
 *<pre>
 * 比如block 的节点,就需要此接口生成对应格式字符
 * block 模块代码 以python 方法的方式呈现
 * 如需方法的方式呈现都可以以此接口表示
 *</pre>
"""
class FuncNode(BaseNode):

    def __init__(self,body,attrs = {}):

        super().__init__(body,attrs)

        return ;


import re
from ..nodes.MainNode import MainNode
from ..utils import ViewUtil


"""
 * 模板编译类
 *<B>说明：</B>
 *<pre>
 *  负责解析模板
 *</pre>
"""
class TemplateCompiler:

    def __init__(self,template = None):

        # 当前的模板对象
        self.template = template

        # 页面的标签列表
        self.pageTags = {}

        # 表达式对象缓存
        self.expressionsCache = [];

        # 默认表达式对象,比如{}
        self.defaultExp = None;

    # 添加标签库
    # <B> 说明： </B>
    # <pre>
    # 模板动态添加标签库
    # </pre>
    def addTags(self,tag:str):

        tagList = tag.split(',')
        for tag in tagList:
            self.pageTags[tag] = tag

        self.buildPageTagExpression();

        return ;

    # 构建当前页面所有的标签表达式对象
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def buildPageTagExpression(self):

        self.expressionsCache = self.template.getTaglibExpression()

        for tagName in self.pageTags:
            self.expressionsCache += self.template.buildTaglibExpression(tagName);

        from .Expression import Expression
        if self.defaultExp is None:
            expAttrs = {'name': "exp", 'close': False, 'outbefore': True,'defaultExp':True, 'handle': self.template.defaultSysTag._exp,
                        'start': self.getExpPattern()}
            self.defaultExp = Expression(expAttrs)

        self.expressionsCache.append(self.defaultExp)

        return ;

    # 获取当前页面的所有标签表达式
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def getAllExpression(self):

        if not self.expressionsCache:
            self.buildPageTagExpression()

        return self.expressionsCache;


    def buildExpTagPattern(self):

        if hasattr(self, "_expTag"):
            return getattr(self,"_expTag");

        self.getAllExpression();
        expPatternList = [];
        for exp in self.expressionsCache:
            if exp.getName() != 'exp':
                if exp.onTag:
                    expPatternList.append(exp.getStart())
            else:
                expPatternList.append(r"{0}".format(exp.getStart()))

        setattr(self,"_expTag",re.compile(r"{0}".format("|".join(expPatternList))))

        return getattr(self,"_expTag");

    def getExpressionPatternList(self):

        self.buildPageTagExpression();
        expressionsList = ViewUtil.split_list(self.expressionsCache,20)

        newExpList = [];
        for expList in expressionsList :
            expPatternList = [];
            for exp in expList:
                expPatternList.append(r"{0}".format(exp.getStart()))

            if expPatternList:
                newExpList.append(r"{0}".format("|".join(expPatternList)))

        return newExpList

    # 获取当前页面所有表达式对象组合的正则表达式对象
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def getAllExpressionPattern(self):

        self.buildPageTagExpression();
        expList = []

        for exp in self.expressionsCache:
            expList.append(r"{0}".format(exp.getStart()))

        return re.compile(r"{0}".format("|".join(expList)))

    def getTemplatePattern(self,pattern = False):

        if self.template.onTag:
            if pattern:
                return re.compile(r"{0}|{1}".format(self.getExpPattern(), self.getTagPattern()))
            else:
                return "{0}|{1}".format(self.getExpPattern(), self.getTagPattern())
        else:
            if pattern:
                return re.compile(r"{0}".format(self.getExpPattern()))
            else:
                return "{0}".format(self.getExpPattern())

    def getExpStart(self):

        return self.template.expStart

    def getExpEnd(self):

        return self.template.expEnd

    def getTagStart(self):

        return self.template.tagStart

    def getTagEnd(self):

        return self.template.tagEnd

    def getTemplateExpReg(self,pattern):

        expReg = r'{0}[^\s].+?{1}'.format(self.template.expStart, self.template.expEnd)
        tagReg = r'{0}[^{1}]*?{2}'.format(self.template.tagStart, self.template.tagEnd, self.template.tagEnd)

        if self.template.onTag:
            if pattern:
                return re.compile(r"({0}|{1})".format(expReg, tagReg))
            else:
                return r"({0}|{1})".format(expReg, tagReg)
        else:
            if pattern:
                return re.compile(r"({0})".format(expReg))
            else:
                return "({0})".format(expReg)

        return re.compile(r"({0}|{1})".format(expReg, tagReg))

    def getExpPattern(self,pattern = False):

        if pattern:
            return re.compile(r'{0}([^\d\s].+?){1}'.format(self.template.expStart, self.template.expEnd));
        else:
            #exp = r'{0}(?![\\n|\s])(.*?){1}'.format(self.template.expStart,self.template.expEnd)
            exp = r'{0}([^\s].+?){1}'.format(self.template.expStart,self.template.expEnd)
            return exp;

    def getTagPattern(self,pattern = False):

        if pattern:
            return re.compile(r'{0}([^{1}]*?){2}'.format(self.template.tagStart, self.template.tagEnd, self.template.tagEnd));
        else:
            return '{0}([^{1}]*?){2}'.format(self.template.tagStart,self.template.tagEnd,self.template.tagEnd);

    # 编译入口
    # <B> 说明： </B>
    # <pre>
    # 此方法预留吧
    # </pre>
    @classmethod
    def parse(cls, content, attrs = {}):

        template = cls(attrs)

        return template.compiler(content)

    # 解析编译模板内容
    # <B> 说明： </B>
    # <pre>
    # 返回模板解析后的内容
    # </pre>
    def compiler(self,content):

        mainNode = MainNode(content,{
            'template':self.template,
            'compiler':self,
        })

        mainNode.start()
        compileContent = mainNode.formatNode();

        return compileContent



from .BaseNode import BaseNode

"""
 * 主节点
 *<B>说明：</B>
 *<pre>
 *  用于节点解析入口，与TemplateCompiler 对象直接对接
 *</pre>
"""
class MainNode(BaseNode):

    def __init__(self,body,attrs = {}):

        super().__init__(body,attrs)

        self.startNode = self;

        return ;

    def start(self):

        funcNode = self.makeFuncNode(self.body)
        self.body = ''
        self.writeFunc('fetch', funcNode);

        return ;



from .base.BaseTemplate import BaseTemplate
from .utils import ViewUtil

"""
 * 模板视图
 *<B>说明：</B>
 *<pre>
 *  主要用于对外提供模板操作接口
 *</pre>
"""
class BaseView:

    def __init__(self,attrs = {}):

        self.data = {}

        # 视图路径
        self.tplPath = '';

        # 是否开启模板缓存
        self.onCache = True;

        # 模板缓存文件目录
        self.cachePath = '';

        # 资源地址,比如js,css,img
        self.urls = {}

        # 模板缓存有效期,单位秒,0 表示无有效期
        self.timeout = 30 * 60;

        self.expStart = '{{';

        self.expEnd = '}}'

        # 表达式结束符号
        self.expEndSign = '/'

        self.onTag = True

        self.tagStart = '<'

        self.tagEnd = '>'
        # 标签结束符号
        self.tagEndSign = '/'

        self.sysTags = [];

        self.customTags = [];

        if attrs:
            ViewUtil.setAttrs(self, attrs)


    def assign(self,name,value):

        self.data[name] = value


    def fetch(self,tplFile,data = {}):

        tpldata = {};
        if data:
            tpldata = data.copy();

        tpldata = dict(tpldata, **self.data);

        self.data = {};
        template = self.getTemplate();

        return template.fetch(tplFile,tpldata)

    def getTemplate(self)->'Template':

        tpl_conf = self.__getTemplateConf()

        return Template(tpl_conf)


    def __getTemplateConf(self):

        return {
            'tplPath':self.tplPath,
            'onCache':self.onCache,
            'cachePath':self.cachePath,
            'urls': self.urls,
            'timeout': self.timeout,
            'expStart': self.expStart,
            'expEnd': self.expEnd,
            'expEndSign': self.expEndSign,
            'onTag': self.onTag,
            'tagStart': self.tagStart,
            'tagEnd': self.tagEnd,
            'tagEndSign': self.tagEndSign,
            'sysTags': self.sysTags,
            'customTags': self.customTags,
        }

class Template(BaseTemplate):

    pass

class View(BaseView):

    pass

# 注册模板过滤器
def reg_temp_filter(funcName = ''):

    def decorator(func):

        name = funcName;
        if not name:
            name = func.__name__

        Template.registerFilter(name,func)

        return func

    return decorator

# 注册模板上下文数据
def reg_temp_context():

    def decorator(func):

        name = func.__name__
        Template.registerContext(name, func)

        return func

    return decorator









from ..view import reg_temp_filter
import time, datetime,re
from decimal import Decimal

# 注入默认过滤器

# 静态资源-地址
@reg_temp_filter('sta')
def sta_filter(self,imgUrl):

    return  '{0}{1}'.format(self.template.urls['static'],imgUrl)

# 静态资源-js
@reg_temp_filter('js')
def js_filter(self,jsUrl):

    return '{0}{1}'.format(self.template.urls['static'],jsUrl)

# 静态资源-css
@reg_temp_filter('css')
def css_filter(self,cssUrl):

    return  '{0}{1}'.format(self.template.urls['static'],cssUrl)

# 静态资源-图片
@reg_temp_filter('img')
def img_filter(self,imgUrl):

    return  '{0}{1}'.format(self.template.urls['static'],imgUrl)

# 外部资源
@reg_temp_filter('res')
def res_filter(self,resUrl):

    return  '{0}{1}'.format(self.template.urls['res'],resUrl)

# 小物件
@reg_temp_filter('widget')
def widget_filter(self,func:str,data = {},options = {}):

    from ..utils import ViewUtil
    pos = func.find(':')
    if pos == -1:
        # 函数
        funcMeta = ViewUtil.getFuncMeta(func)
        return funcMeta(data)
    else:
        # 类方法
        clazz = func[0:pos]
        clazzMeta = ViewUtil.getModuleMeta(clazz)
        funcName = func[pos + 1:]
        funcMeta = getattr(clazzMeta(data),funcName)

        return funcMeta()

# 计算字符长度
@reg_temp_filter('len')
def len_filter(value):

    text = "{0}".format(value)

    return len(text)

# 去掉两边空格
@reg_temp_filter('trim')
def trim_filter(value):

    text = "{0}".format(value)

    return text.strip()

# 去掉左边空格
@reg_temp_filter('ltrim')
def ltrim_filter(value):

    text = "{0}".format(value)

    return text.lstrip()

# 去掉右边空格
@reg_temp_filter('rtrim')
def rtrim_filter(value):

    text = "{0}".format(value)

    return text.rstrip()

# 日期格式化
@reg_temp_filter('date')
def date_filter(datestr,format):
    if datestr is None:
        now = datetime.datetime.now()
        return  now.strftime(format)
    else:

        if datestr:
            if isinstance(datestr,int):
                #时间戳
                timeItems = time.localtime(datestr)
            else:
                timeItems = time.strptime(datestr, "%Y-%m-%d %H:%M:%S")
            return time.strftime(format, timeItems)
        else:
            return ''
    pass


# 截取字符串
@reg_temp_filter('substr')
def substr_filter(value:str,start_pos = 0,sublen = 0):

    if sublen:
        return value[start_pos:sublen]
    else:
        return value[start_pos:]

# 浮点型,保留小数点
@reg_temp_filter('float')
def float_filter(value:float,point = None):
    decimal = Decimal(value)
    if  point is None:
        return round(decimal)

    if isinstance(point,int):
        return round(decimal,point)

    return value

# 去除html 标签
@reg_temp_filter('striptags')
def striptags_filter(html:str):

    regex = re.compile(r'<[^>]+>', re.S)
    html = regex.sub('', html)

    return html



import re
from ..utils import ViewUtil

"""
 * 节点基类
 *<B>说明：</B>
 *<pre>
 *  
 *</pre>
"""
class BaseNode:

    NODE_TYPE_STR = 1;# 字符串
    NODE_TYPE_CODE = 2;# py 代码
    NODE_TYPE_NODE = 3;# 子节点
    NODE_TYPE_SAME = 4;  # 同级别

    def __init__(self,body:str,attrs = {}):

        self.body = body
        self.template = None
        self.startNode = None
        self.nodes = [];
        self.compiler = None;
        self.findpos = 0
        self.posend = 0
        # 停止查找
        self.stop = False
        # 表达式匹配的位置
        self.outstartpos = 0
        self.outendpos = 0

        # 格式化参数
        self.formatParams = [];

        if attrs:
            ViewUtil.setAttrs(self,attrs)

        return ;

    def getBody(self):

        return self.body

    def getLineBody(self):

        return self.body.split(r'\n')


    def writeBeforeBody(self,all = False):

        if all is True:
            beforeBody = self.body[self.outstartpos:]
        else:
            beforeBody = self.body[self.outstartpos:self.outendpos]

        beforeBody = re.sub('"', r'\"', beforeBody);

        if self.formatParams:
            #outBeforeBody = re.sub(re.compile(r"{0}".format(self.compiler.getExpPattern())), lambda x: "{" + addnumber() + "}", beforeBody)
            beforeBody = re.sub('%', r'%%', beforeBody);
            outBeforeBody = re.sub(re.compile(r"{0}".format(self.compiler.getExpPattern())),'%s', beforeBody)
        else:
            outBeforeBody = re.sub(re.compile(r"{0}".format(self.compiler.getExpPattern())), "", beforeBody)

        if self.formatParams:
            outBeforeBody = '"{0}" % ({1})'.format(outBeforeBody,','.join(self.formatParams))
        else:
            outBeforeBody = '"{0}" '.format(outBeforeBody)

        self.formatParams = []
        self.writeRawCode(outBeforeBody,True,True)
        self.outstartpos = self.outendpos

        return ;

    def addFormatParams(self,name):

        self.formatParams.append(name);

        return ;

    # 导入标签库
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def loadTaglib(self,taglib):

        self.compiler.addTags(taglib)

        return ;

    def find(self):

        #regexbody = self.compiler.getTemplateExpReg(True)

        self.posend = 0
        expList = self.compiler.getExpressionPatternList();
        regexbody = self.compiler.buildExpTagPattern();

        while True:

            if self.stop:
                break;

            resbody = regexbody.search(self.body, self.findpos)
            if resbody is None:
                self.writeBeforeBody(True)
                self.posend = 0
                break

            ratternResult = False;
            for regexStr in expList:
                regex = re.compile(regexStr);
                res = regex.match(self.body, resbody.span()[0])

                if res is None:
                    continue;

                ratternResult = True;
                expPos = self.findFullExp(res);
                resStartPos = expPos[0]
                resEndPos = expPos[1]
                expRawContent = self.body[resStartPos:resEndPos]

                expContent = expPos[2]
                self.posend = resEndPos

                # 此处,可优化性能
                self.parseExp(expRawContent,expContent,resStartPos,resEndPos);
                self.findpos = self.posend

                break;

            if not ratternResult:
                self.findpos = resbody.span()[1]

        return ;

    def parseExp(self,expRawContent,expContent,resStartPos,resEndPos):

        # 此处,可优化性能
        expressionList = self.compiler.getAllExpression()
        for exp in expressionList:
            if self.stop:
                break;
            startCompile = exp.getStartRule()
            startMatch = startCompile.search(expRawContent)

            if startMatch is not None:
                handleFunc = exp.getHandle();
                self.outendpos = resStartPos;
                # 读取查找之前的内容
                if exp.getOutbefore() is False:
                    self.writeBeforeBody()

                tagAttrs = startMatch.group(1)
                if exp.hasEnd():
                    # 标签类型
                    spanPos = self.findNodeBodyPos(self.posend, exp)
                    if spanPos:
                        self.posend = spanPos[1]
                        tagBody = self.body[resEndPos:spanPos[0]]
                        subTagNode = self.makeNode(tagBody)
                        result = handleFunc(tagAttrs, self, subTagNode)

                        if result is False:
                            break
                else:
                    # {} 类型
                    if exp.hasdefaultExp():
                        result = handleFunc(expContent, self)
                    else:
                        result = handleFunc(tagAttrs, self)

                    if result is False:
                        break

                if exp.getOutbefore() is False:
                    self.outstartpos = self.posend

                break

        return ;


    # 匹配完整的表达式
    # <B> 说明： </B>
    # <pre>
    # 由于{} 存在嵌套的情况,需二次查找
    # </pre>
    def findFullExp(self,pattern):

        # 判断起始符号 是表达式或标签
        expStart = self.compiler.getExpStart();
        tagStart = self.compiler.getTagStart();
        nodeContent = pattern.group()
        nodeStartChar = "";
        nodeEndChar = ""

        if expStart == nodeContent[0:len(expStart)]:
            nodeStartChar = expStart
            nodeEndChar = self.compiler.getExpEnd();
            pass
        elif tagStart == nodeContent[0:len(tagStart)]:
            nodeStartChar = tagStart
            nodeEndChar = self.compiler.getTagEnd();
            pass

        # 计算位置
        startNum = nodeContent.count(nodeStartChar)
        endCharNum = startNum - 1;
        expstartpos = pattern.span()[0]
        find_pos = pattern.span()[1];
        expendpos = pattern.span()[1];
        if endCharNum > 0:
            while (True):
                find_pos = self.body.find(nodeEndChar,find_pos)
                if find_pos == -1:
                    break

                if find_pos != -1:
                    endCharNum -= 1;
                    expendpos = find_pos;

                if endCharNum <= 0:
                    break;
            expendpos += 1

        return (expstartpos,expendpos,self.body[expstartpos + len(nodeStartChar):expendpos - len(nodeEndChar)])

    # 终端当前节点查找
    # <B> 说明： </B>
    # <pre>
    # 即可终止find 的查找匹配操作
    # </pre>
    def stopFind(self):

        self.stop = True

        return ;

    def makeNode(self,tagBody,format = False):

        from .TagNode import TagNode

        if format:
            lines = tagBody.splitlines()
            tagBody = r'\n'.join(lines)

        tagNode = TagNode(tagBody,
                {'template': self.template,
                 'compiler': self.compiler,
                 'startNode':self.startNode
                 });

        return tagNode

    def makeFuncNode(self,tagBody):

        from .FuncNode import FuncNode
        funcNode = FuncNode(tagBody,
                          {'template': self.template,
                           'compiler': self.compiler,
                           'startNode': self.startNode
                           });

        return funcNode

    def getNofindBody(self):

        if self.posend == 0:
            return self.body
        else:
            return  self.body[self.posend:]

    # 通过入栈的方式匹配表达式成对的起始符号与结束符号
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def findNodeBodyPos(self,findpos,exp):

        expNum = 1
        spanPos = ();
        while True:
            if expNum <= 0:
                break;

            startEndCompile = exp.getStartEndRule()
            startEndMatch = startEndCompile.search(self.body, findpos)

            if startEndMatch is not None:
                endCompile = exp.getEndRule()
                findpos = startEndMatch.span()[1]
                if endCompile.search(startEndMatch.group()) is not None:
                    # 找到结束符
                    expNum = expNum - 1
                    spanPos = startEndMatch.span()
                else:
                    expNum = expNum + 1


        return spanPos

    # 输出当前节点
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def formatNode(self,depth = 1):

        nodeList = []
        for (nodeType,node,tab,yld) in self.nodes:
            nodeStr = ''
            if nodeType == self.NODE_TYPE_STR:
                nodeStr = node
                if tab:
                    nodeStr = ('    ' * depth) + nodeStr
                else:
                    nodeStr = ('    ' * (depth - 1)) + nodeStr
            elif nodeType == self.NODE_TYPE_CODE:
                nodeStr = node
                if tab:
                    nodeStr = ('    ' * depth) + nodeStr
                else:
                    nodeStr = ('    ' * (depth - 1)) + nodeStr
            elif nodeType == self.NODE_TYPE_NODE:
                nodeStr = node.formatNode(depth + 1)
            elif nodeType == self.NODE_TYPE_SAME:
                nodeStr = node.formatNode(depth)

            if nodeStr:
                nodeList.append(nodeStr)

        return '\n'.join(nodeList)

    # 写入节点对象
    def writeNode(self, node, tab = True,yld = False):

        self.nodes.append((self.NODE_TYPE_NODE,node,tab,yld))

        return ;

    def writeSameNode(self, node, tab = True,yld = False):

        self.nodes.append((self.NODE_TYPE_SAME,node,tab,yld))

        return ;



    # 写入代码
    def writeCode(self,code,tab = True,yld = False):

        if yld:
            code = 'yield "%s" % ({0})'.format(code)

        self.nodes.append((self.NODE_TYPE_CODE,code,tab,yld))

        return ;

    def writeRawCode(self,code,tab = True,yld = False):

        if yld:
            code = 'yield {0}'.format(code)

        self.nodes.append((self.NODE_TYPE_CODE,code,tab,yld))

        return ;

    # 写入字符
    def writeStr(self,chars,tab = True,yld = True):

        if yld:
            chars = re.sub('"', r'\"', chars);
            chars = 'yield "{0}"'.format(chars)

        self.nodes.append((self.NODE_TYPE_STR, chars,tab,yld))

        return ;

    # 写入方法
    def writeFunc(self,funcName,bodyNode):

        self.writeCode('def {0}(self):'.format(funcName),True,False)
        self.writeNode(bodyNode,False,False)
        bodyNode.find();

        return ;






from .Tag import Tag
from ..nodes.BaseNode import BaseNode

"""
 * html标签库
 *<B>说明：</B>
 *<pre>
 *  略
 *</pre>
"""
class HtmlTag(Tag):

    def __init__(self,template):

        super().__init__(template)
        self.tags = [
            {'name': "css", 'close': False, 'onTag': True},
            {'name': "js", 'close': False, 'onTag': True},
            {'name': "img", 'close': False, 'onTag': True},
        ]

        return ;

    def _css(self,node_attrs, mainNode:BaseNode):
        attrs = self.parseAttr(node_attrs)
        href = attrs['href'];
        cssFunc = 'self.call("css","{0}")'.format(href)
        attrs['href'] = '%s';
        attrsHtml = self.buildHtmlAttrs(attrs);
        codeHtml = "'<link {0} />' % ({1})".format(attrsHtml, cssFunc)

        mainNode.writeRawCode(codeHtml, True, True)

    def _js(self,node_attrs, mainNode:BaseNode):
        attrs = self.parseAttr(node_attrs)
        src = attrs['src'];
        jsFunc = 'self.call("js","{0}")'.format(src)
        attrs['src'] = '%s';
        attrsHtml = self.buildHtmlAttrs(attrs);
        codeHtml = "'<script {0} ></script>' % ({1})".format(attrsHtml, jsFunc)

        mainNode.writeRawCode(codeHtml, True, True)

    def _img(self,node_attrs, mainNode:BaseNode):
        attrs = self.parseAttr(node_attrs)
        src = attrs['src'];
        jsFunc = 'self.call("img","{0}")'.format(src)
        attrs['src'] = '%s';
        attrsHtml = self.buildHtmlAttrs(attrs);
        codeHtml = "'<img {0} >' % ({1})".format(attrsHtml, jsFunc)

        mainNode.writeRawCode(codeHtml, True, True)












from .BaseNode import BaseNode

"""
 * 普通标签节点
 *<B>说明：</B>
 *<pre>
 *  略
 *</pre>
"""
class TagNode(BaseNode):

    def __init__(self,body,attrs = {}):

        super().__init__(body,attrs)

        return ;


from .Tag import Tag
from ..nodes.BaseNode import BaseNode

"""
 * html标签库
 *<B>说明：</B>
 *<pre>
 *  使用
 *</pre>
"""
class FormTag(Tag):

    def __init__(self,template):

        super().__init__(template)
        self.tags = [
            {'name': "select", 'close': False,'onTag':True},
            {'name': "text", 'close': False, 'onTag': True},
            {'name': "radio", 'close': False, 'onTag': True},
            {'name': "checkbox", 'close': False, 'onTag': True},
        ]

        return ;


    # select 标签
    def _select(self,node_attrs,mainNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        htmlAttrs,heAttrs = self.spiltAttr(attrs);
        hename = heAttrs.get("he-name")
        hetext = heAttrs.get("he-text","")
        hevalue = heAttrs.get("he-value", None)

        if not hevalue:
            hevalue = "_request.{0}".format(attrs['name'])

        attrsHtml = self.buildHtmlAttrs(htmlAttrs);
        html = '''<select {0}>
               <option value="">{1}</option>
               <for name="{2}" key="_key" value="_value">
                   <if condition="str(_key) == str({3})">
                        <option selected="selected" value="{4}">{5}</option>
                   <else/>
                        <option value="{4}">{5}</option>
                   </if>
               </for>
               </select>'''

        html = html.format(attrsHtml,hetext,hename,hevalue,self.buildVar("_key"),self.buildVar("_value"));
        tagNode = mainNode.makeNode(html,True)
        tagNode.find();

        mainNode.writeSameNode(tagNode,False,False)

        pass

    # input text 标签
    def _text(self, node_attrs, mainNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        htmlAttrs, heAttrs = self.spiltAttr(attrs);
        hevalue = heAttrs.get("he-value",None)

        if not hevalue:
            hevalue = self.buildVar("_request.{0}".format(attrs['name']))
        else:
            hevalue = self.buildVar(hevalue)

        attrsHtml = self.buildHtmlAttrs(htmlAttrs);
        html = '''<input type="text" {0} value="{1}" />'''

        html = html.format(attrsHtml, hevalue);
        tagNode = mainNode.makeNode(html, True)
        tagNode.find();

        mainNode.writeSameNode(tagNode, False, False)

        pass

    # input text 标签
    def _radio(self, node_attrs, mainNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        htmlAttrs, heAttrs = self.spiltAttr(attrs);
        hevalue = heAttrs.get("he-value", None)
        radiovalue = htmlAttrs.get("value", '')

        if not hevalue:
            hevalue = "_request.{0}".format(attrs['name'])

        attrsHtml = self.buildHtmlAttrs(htmlAttrs);
        html = '''<input type="radio" {0} <if condition="str({1}) == str({2})">checked="checked"</if> />'''
        html = html.format(attrsHtml, hevalue,radiovalue);
        tagNode = mainNode.makeNode(html, True)
        tagNode.find();

        mainNode.writeSameNode(tagNode, False, False)

        pass

    # input text 标签
    def _checkbox(self, node_attrs, mainNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        htmlAttrs, heAttrs = self.spiltAttr(attrs);
        hevalue = heAttrs.get("he-value", None)
        radiovalue = htmlAttrs.get("value", '')

        if not hevalue:
            hevalue = "_request.{0}".format(attrs['name'])

        attrsHtml = self.buildHtmlAttrs(htmlAttrs);
        html = '''<input type="checkbox" {0} <if condition="str({1}) == str({2})">checked="checked"</if> />'''
        html = html.format(attrsHtml, hevalue, radiovalue);
        tagNode = mainNode.makeNode(html, True)
        tagNode.find();

        mainNode.writeSameNode(tagNode, False, False)

        pass
















from ..base.Expression import Expression
from ..utils import ViewUtil
import re

"""
 * 标签库基类
 *<B>说明：</B>
 *<pre>
 *  略
 *</pre>
"""
class Tag:

    def __init__(self,template):

        self.alias = ''
        self.tags = []
        self.template = template

        return ;

    def buildVar(self,name):

        var = '{0}{1}{2}'.format(self.template.expStart,name,self.template.expEnd)

        return var;

    @classmethod
    def make(cls,tagName,template)->'Tag':
        if not tagName:
            raise RuntimeError('tagName is empty')

        if tagName.find('.') == -1:
            tagClazzName = ViewUtil.ucfirst(tagName) + 'Tag'
            tagClazz = __package__ + "." + tagClazzName + '.' + tagClazzName
        else:
            tagClazz = tagName

        tagMeta = ViewUtil.getModuleMeta(tagClazz)

        return tagMeta(template)

    def getAlias(self):

        return self.alias

    def getTagExpression(self,alias = ''):
        expList = [];

        for tag in self.tags:
            attrs = tag.copy()

            tagName = attrs['name']
            if alias:
                attrs['name'] = '{0}:{1}'.format(alias,tagName)

            # 处理函数
            handle = tag.get('handle', None)
            if handle is None:
                handle = '_{0}'.format(tagName)

            attrs['handle'] = getattr(self,handle)

            # 起始符号
            startTag = tag.get('start', None)
            if startTag is None:
                attrs['start'] = r'{0}{1}(.*?){2}'.format(self.template.expStart, attrs['name'], self.template.expEnd)
            else:
                attrs['start'] = tag['start']

            # 结束符号符号
            closeTag = tag.get('close',None)
            if closeTag:
                eofName = tag.get('eof', self.template.expEndSign)
                attrs['end'] = '{0}{1}{2}{3}'.format(self.template.expStart,eofName,attrs['name'],self.template.expEnd)

            expList.append(Expression(attrs))

            # 启用标签
            onTag = tag.get('onTag', None)
            if onTag:
                tagAttrs = attrs.copy()
                # 起始符号
                tagStart = '{0}{1}([^a-zA-Z0-9].*?){2}'.format(self.template.tagStart, tagAttrs['name'], self.template.tagEnd)
                tagAttrs['start'] = r'{0}'.format(tagStart)

                # 结束符号符号
                closeTag = tag.get('close', None)
                if closeTag:
                    eofName = tag.get('eof', self.template.tagEndSign)
                    tagEnd = '{0}{1}{2}{3}'.format(self.template.tagStart, eofName, tagAttrs['name'],self.template.tagEnd)
                    tagAttrs['end'] = r'{0}'.format(tagEnd)

                expList.append(Expression(tagAttrs))

        return expList

    # 构建标签表达式
    def buildTagExp(self,exp:str):

        return exp.format(self.template.expStart,self.template.expStart)

    def buildCommonExp(self,exp:str):

        return exp.format(self.template.expStart,self.template.expEnd)

    # 解析变量{name} {:U('xxx')}
    def parseVar(self,varBody):
        if not varBody:
            return ''

        # 读取首字母
        firstChar =  varBody[0]
        result = '';
        if  firstChar == '$':
            # 变量
            result = self.parseName(varBody[1:],True)
        elif firstChar == ':':
            #方法
            func = varBody[1:];
            funcFirstChar = func[0];
            funcList = func.split('|')
            findpos = funcList[0].find('(')
            if findpos == -1:
                if funcFirstChar == '$':
                    funcList[0] = self.parseName(funcList[0][1:],True);
                else:
                    funcList[0] = self.parseName(funcList[0],False);

            func = "|".join(funcList)
            result = self.parseFunc(func)
        else:
            # 不处理,直接返回
            result = self.parseName(varBody);

        return result

    def parseName(self,varname,sale = False):

        #regex = re.compile("\['(\\w+)'\]|" + '\["(\\w+)"\]')
        regex1 = re.compile("\['(\\w+)'\]")
        regex2 = re.compile('\["(\\w+)"\]')
        matches1 = regex1.findall(varname);
        matches2 = regex2.findall(varname);
        if matches1 or matches2:

            keys = [];
            for key in matches1:
                keys.append(key);

            for key in matches2:
                keys.append(key);

            varnamekey = varname[0:varname.find('[')];
            if sale :
                newName = 'self.getvalue({0},"{1}",True)'.format(varnamekey,".".join(keys));
            else:
                newName = 'self.getvalue({0},"{1}")'.format(varnamekey, ".".join(keys));
            return newName
        else:
            return varname;

    # 解析属性 name="admin" file="yonghu"
    def parseAttr(self,attrBody):

        if not attrBody:
            return attrBody

        if isinstance(attrBody,dict):
            return attrBody
        reg = r'\s+([\-a-zA-Z]+)\s*=\s*"([^"]+)"'
        pattern = re.compile(reg);

        matchs = re.findall(pattern, attrBody)
        if not matchs:
            return attrBody

        attrs = {}
        for match in matchs:  # i 本身也是可迭代对象，所以下面要使用 i.group()
            attrs[match[0]] = match[1]

        return attrs


    def parseNode(self,mainNode,subNode):

        subNode.find()
        mainNode.writeNode(subNode)

        return ;


    # 解析方法
    def parseFunc(self,func:str):

        funcList = func.split('|')
        findpos = funcList[0].find('(')
        if findpos == -1:
            value = funcList.pop(0)
        else:
            value = '';

        for funcstr in funcList:
            funcName,funcParams = self.formatFunc(funcstr);
            value = self.buildFuncCall(funcName,funcParams,value);

        return value

    def buildFuncCall(self,funcName,funcParams:str,value):

        # 验证过滤器存在
        if self.template.hasfilters(funcName):
            if value:
                return 'self.call("{0}",{1})'.format(funcName, funcParams.format(value))
            else:
                return 'self.call("{0}",{1})'.format(funcName, funcParams)
        else:
            if value:
                return '{0}({1})'.format(funcName, funcParams.format(value))
            else:
                return '{0}({1})'.format(funcName, funcParams)



    # 分析字符串方法
    def formatFunc(self,func:str):

        startPos = func.find('(')
        if startPos == -1:
            # 未找到(
            paramsPos = func.find('=')
            if paramsPos == -1:
                funcName = func
                funcParams = '###';
            else:
                funcName = func[0:paramsPos]
                funcParams = func[paramsPos + 1:]
                if funcParams.find("###") == -1:
                    funcParams = '###,{0}'.format(funcParams)

        else:
            endPos = func.rfind(')')
            funcName = func[0:startPos]
            funcParams = func[startPos + 1:endPos]

        funcParams = funcParams.replace('###', "{0}")

        return funcName,funcParams

    def buildBlockFuncName(self,funcName):

        return 'layout_block_{0}'.format(funcName)

    def buildBlockFuncList(self,funcs:dict):

        blockList = {}
        for funcName,funcContent in funcs.items():
            blockList[self.buildBlockFuncName(funcName)] = funcContent

        return blockList

    def parseBlock(self,block):

        blockList = {}

        if block is None:
            blockList["default"] = "__CONTENT__"
        else:
            blocks = str(block).split(',')
            for block in blocks:
                blockValue = block.split(":")
                blockList[blockValue[0]] = blockValue[1]

        blockList = self.buildBlockFuncList(blockList)

        return blockList

    def buildHtmlAttrs(self,attrs = {}):

        html = []

        for name,value in attrs.items():
            html.append('{0}="{1}"'.format(name,value))

        return " ".join(html)


    def spiltAttr(self,attrs:dict):

        htmlAttrs = {};
        heAttrs = {};
        for attrname,attrvalue in attrs.items():
            if attrname[0:3] == 'he-':
                heAttrs[attrname] = attrvalue
            else:
                htmlAttrs[attrname] = attrvalue

        return [htmlAttrs,heAttrs];




from ..view import reg_temp_filter
"""
 * 安全过滤器
 *<B>说明：</B>
 *<pre>
 * 略
 *</pre>
"""

html_xss = {
    '&':'&amp;',
    '>':'&gt;',
    '<':'&lt;',
    "'":'&#39;',
    '"':'&#34;',
}

# xss 过滤器
@reg_temp_filter('xss')
def xss_filter(value):
    text = '{0}'.format(value)
    for find,replace in html_xss.items():
        text = text.replace(find,replace)

    return text

# 过滤所有有风险的字符
@reg_temp_filter('safe')
def safe_filter(self,value):

    value = xss_filter(value)

    return value



from ..utils import ViewUtil
import re

"""
 * 表达式类
 *<B>说明：</B>
 *<pre>
 *  略
 *</pre>
"""
class Expression:

    def __init__(self,attrs = []):

        # 起始符号
        self.start = None;
        # 结束符号
        self.end = None;
        # 表达式名称,比如for,if list
        self.name = None
        # 表达式对应的处理方法,表达式匹配成功后,交由此方法处理
        self.handle = None
        # 是否输出之前的字符
        self.outbefore = False
        # 是否默认表达式({})
        self.defaultExp = False;
        # 是否使用标签
        self.onTag = False;

        if attrs:
            ViewUtil.setAttrs(self,attrs)

    def getName(self):

        return self.name

    def getOutbefore(self):

        return self.outbefore

    def hasdefaultExp(self):

        return self.defaultExp

    def getHandle(self):

        return self.handle

    def hasEnd(self):

        if self.end:
            return True
        else:
            return False

    def getStart(self):

        return self.start

    def getEnd(self):

        return self.end

    def getEndRule(self):

        return re.compile(r'{0}'.format(self.end));

    def getStartRule(self):
        return re.compile(r'{0}'.format(self.start));

    def getStartEndRule(self):

        return  re.compile(r'{0}|{1}'.format(self.start,self.end))







"""
 * 模板上下文
 *<B>说明：</B>
 *<pre>
 *  使用
 *</pre>
"""
class TemplateContext:

    def __init__(self):

        self.template = None

        self.contextList = {} # 上下文定义

        return ;


    def register(self,funcName, funcMeta):

        self.contextList[funcName] = funcMeta

        return ;

    def getContext(self)->'list':

        data = {};
        for funcName,funcMeta in self.contextList.items():
            data.update(funcMeta())

        return data





"""
 * 过滤器类
 *<B>说明：</B>
 *<pre>
 *  模板的调用的方法必须注入到此对象
 *</pre>
"""
class Filters:

    def __init__(self):

        self.template = None

        return ;


    def setTemplate(self,template):

        self.template = template

        return ;





"""
 * 缓存模板基类
 *<B>说明：</B>
 *<pre>
 *  略
 *</pre>
"""
class MyTemplate:

    def __init__(self,template = None):

        self.template = template

        self._tempdata = None;

        return ;

    # 模板方法调用统一入口
    # <B> 说明： </B>
    # <pre>
    # 通过此方法间接调用过滤器的方法
    # </pre>
    def call(self,funcName,*params):

        funcMeta = getattr(self.template.__class__.filters, funcName)

        return funcMeta(*params)

    # layout 标签辅助方法
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def layout(self,tplFile,blocks:dict):

        layoutContent = str(self.template.loadTemplate(tplFile))

        for blockName,blockContent in blocks.items():
            blockFunc = getattr(self,blockName)
            layoutContent = layoutContent.replace(blockContent, ''.join(blockFunc()))

        return layoutContent

    # block 标签辅助方法
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def block(self,blockName):

        blockFunc = getattr(self, blockName)

        return ''.join(blockFunc());

    # include 标签辅助方法
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def include(self,tplFile):

        return self.template.loadTemplate(tplFile)

    # for 标签辅助方法
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def fordata(self,data,attrs = {}):

        var_key = attrs.get('key', None)
        var_value = attrs.get('value', None)

        if isinstance(data,list):
            if  var_key is None:
                return data;
            else:
                if var_value is None:
                    return range(len(data))
                else:
                    return enumerate(data)

        elif isinstance(data,dict):
            if var_value is not None and var_key is not None:
                return data.items()
            else:
                return data

        return data

    def getvalue(self,valueDic,keys:str,sale = False):

        if  not valueDic:
            return valueDic;

        keyList = keys.split(".")
        tempVal = valueDic;
        for key in keyList:
            if not tempVal:
                return ;
            tempVal = tempVal.get(key,'')

        if sale :
            tempVal = self.call("safe",tempVal)

        return tempVal;

    # for 标签辅助方法
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    def fortag(self, data):

        if isinstance(data, list):
            return data;
        elif isinstance(data, dict):
            data.items()

        return data


# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open("./README.md", "r",encoding="utf-8") as fh:
    long_description = fh.read()

#python3.6 setup.py sdist upload

setup(
    name = 'hehey-htemplate',
    version = '1.0.2',
    author = 'hehe',
    packages=find_packages(),
    author_email = 'chinabluexfw@163.com',
    url = 'https://gitee.com/chinahehe/hehey-htemplate',
    description = 'hehey-htemplate 是一个python 轻量的模板引擎,其主要特点有:易学,示例全,功能全面,html友好标签,编译速度快,易扩展与其他模板引擎对比,其简单易学,速度快(大概1000次编译,800 多毫秒),随时随地编写自己的标签库.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    #package_dir = 'htemplate'
)



import importlib,re,os

class ViewUtil:

    # 获取类或对象的自定义属性
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    @classmethod
    def setAttrs(cls, object, attrDict={}):

        for attr in attrDict:
            setattr(object, attr, attrDict[attr])

    @classmethod
    def ucfirst(self, str):

        return str[0].upper() + str[1:]

    # 获取上一级目录
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    @classmethod
    def dirname(cls, clazz):

        packageClass = clazz.rsplit('.', 1)

        return packageClass[0];

    # 获取类或对象的自定义属性
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    @classmethod
    def getModuleMeta(cls, clazz):

        if type(clazz) == str:
            packageClass = clazz.rsplit('.', 1)
            packageMeta = importlib.import_module(packageClass[0])
            return getattr(packageMeta, packageClass[1])
        else:
            return clazz

    # 获取指定文件的扩展名
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    @classmethod
    def getFileExt(self,filename):

        filelist = os.path.splitext(filename)
        ext = filelist[len(filelist) - 1]
        if ext:
            ext = ext[1:]

        return ext

    # 获取类或对象的自定义属性
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    @classmethod
    def getFuncMeta(cls, func):

        if type(func) == str:
            packageFunc = func.rsplit('.', 1)
            packageMeta = importlib.import_module(packageFunc[0])
            return getattr(packageMeta, packageFunc[1])
        else:
            return func

    # 统计左边空格数量
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    @classmethod
    def countlstrip(cls,chars):

        reg = '^(\s*)\w+'
        result = re.match(reg,chars)
        if result:
            count = result.group(1).count(' ')
            return count
        else:
            return False

    # 分割list
    # <B> 说明： </B>
    # <pre>
    # 略
    # </pre>
    @classmethod
    def split_list(cls,splitlist,n):

        i = 0;
        newlist = [];
        templist = [];
        for value in splitlist:
            templist.append(value);

            if i >= n:
                newlist.append(templist.copy());
                templist = [];
                i = 0;

            i = i + 1

        if templist:
            newlist.append(templist.copy());

        return newlist





from ..utils import ViewUtil
from .Tag import Tag
from ..nodes.BaseNode import BaseNode

"""
 * 系统标签库
 *<B>说明：</B>
 *<pre>
 *  略
 *</pre>
"""
class SysTag(Tag):

    def __init__(self,template):

        super().__init__(template)
        self.tags = [
            {'name': "taglib", 'close': False},# 标签库
            {'name': "layout", 'close': False},
            {'name': "block", 'close': True},
            {'name': "call", 'close': False},# 调用模板内部方法
            {'name': "include", 'close': False},
            {'name': "import", 'close': False},# 导入python 模块
            {'name': "css" ,'close': False,'onTag':True},
            {'name': "js", 'close': False, 'onTag': True},
            {'name': "for", 'close': True,'onTag':True},
            {'name': "raw", 'close': False},
            {'name': "list", 'close': True,'onTag':True},
            {'name': "dict", 'close': True,'onTag':True},
            {'name': "if", 'close': True,'onTag':True},
            {'name': "else", 'close': False,'onTag':True},
            {'name': "elif", 'close': False,'onTag':True},
            {'name': "python", 'close': True},# 原始python 源码
            {'name': "comment", 'close': True},  # 注释标签
            {'name': "pass", 'close': True},  # 不解析标签
            {'name': "set", 'close': False},  # 定义变量
            {'name': "empty", 'close': True},  # 判断空
            {'name': "notempty", 'close': True},  # 判断不未空
        ]

        return ;

    # {}
    def _exp(self,node_attrs,mainNode:BaseNode):

        expStr = self.parseVar(node_attrs)
        mainNode.addFormatParams(expStr)
        pass

    # 模板继承
    def _layout(self,node_attrs,mainNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        fileName =  attrs['file']
        block = attrs.get("block",None)
        blockList = self.parseBlock(block)
        mainNode.writeCode('self.layout("{0}",{1})'.format(fileName, blockList), True, True)

        if block is None:
            tagNode = mainNode.makeNode(mainNode.getNofindBody())
            mainNode.startNode.writeFunc(self.buildBlockFuncName('default'), tagNode)
            mainNode.stopFind()


        return False

    # 加载文件
    def _include(self,node_attrs,mainNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        fileName = attrs['file']
        mainNode.writeCode('self.include("{0}")'.format(fileName), True, True)
        pass

    # 加载python 模块
    def _import(self,node_attrs,mainNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        name = attrs['name']
        alias = attrs.get("as")
        pos = name.find(':')
        if pos == -1:
            mainNode.writeCode('import {0}'.format(name), True)
        else:
            packageMeta = name.rsplit(':', 1)
            if alias :
                mainNode.writeCode('from {0} import {1} as {2}'.format(packageMeta[0],packageMeta[1],alias), True)
            else:
                mainNode.writeCode('from {0} import {1}'.format(packageMeta[0], packageMeta[1]), True)

        pass

    # 块模块
    def _block(self, node_attrs, mainNode:BaseNode,subNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        blockName = attrs['name']
        mainNode.startNode.writeFunc(self.buildBlockFuncName(blockName), subNode)

        pass

    # 页面导入标签库
    def _call(self, node_attrs, mainNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        if not node_attrs:
            return

        blockName = attrs['name']
        mainNode.writeCode('self.block("{0}")'.format(self.buildBlockFuncName(blockName)), True, True)

        pass

    # 页面导入标签库
    def _taglib(self, node_attrs, mainNode:BaseNode):

        if not node_attrs:
            return

        attrs = self.parseAttr(node_attrs)
        if not isinstance(attrs, dict):
            mainNode.loadTaglib(node_attrs.strip())
        else:
            mainNode.loadTaglib(attrs['name'])

        return False

    def _css(self,node_attrs, mainNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        href = attrs['href'];
        cssFunc = 'self.call("css","{0}")'.format(href)
        attrs['href'] = '%s';
        attrsHtml = self.buildHtmlAttrs(attrs);
        codeHtml = "'<link {0} />' % ({1})".format(attrsHtml, cssFunc)

        mainNode.writeRawCode(codeHtml, True, True)

    def _js(self,node_attrs, mainNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        src = attrs['src'];
        jsFunc = 'self.call("js","{0}")'.format(src)
        attrs['src'] = '%s';
        attrsHtml = self.buildHtmlAttrs(attrs);
        codeHtml = "'<script {0} ></script>' % ({1})".format(attrsHtml, jsFunc)

        mainNode.writeRawCode(codeHtml, True, True)

    def _for(self,node_attrs,mainNode:BaseNode,subNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        if  not isinstance(attrs,dict):
            mainNode.writeCode('for {0}:'.format(node_attrs))
            self.parseNode(mainNode, subNode)
            return

        var_data = attrs['name']
        var_key = attrs.get('key', None)
        var_value = attrs.get('value', None)
        var_index = attrs.get('index', None)

        if var_index is not None:
            mainNode.writeCode('{0} = 0'.format(var_index), True, False)

        if var_key is None:
            mainNode.writeCode('for {0} in self.fordata({1},{2}):'.format(self.parseVar(var_value), self.parseVar(var_data),attrs))
        else:
            mainNode.writeCode(
                'for {0},{1} in self.fordata({2},{3}):'.format(self.parseVar(var_key), self.parseVar(var_value),
                                                        self.parseVar(var_data),attrs))
        # 继续解析node body
        self.parseNode(mainNode,subNode)

        if var_index is not None:
            subNode.writeCode('{0}+=1'.format(var_index),True,False)


    def _list(self,node_attrs,mainNode:BaseNode,subNode:BaseNode):

        attrs = self.parseAttr(node_attrs)
        if not isinstance(attrs, dict):
            mainNode.writeCode('for {0}:'.format(node_attrs))
            self.parseNode(mainNode, subNode)
            return

        varData = attrs['name']
        varKey = attrs.get('key',None)
        varValue = attrs.get('value', None)
        varIndex = attrs.get('index', None)

        if varIndex is not None:
            mainNode.writeCode('{0} = 0'.format(varIndex), True, False)

        if varKey is None:
            mainNode.writeCode('for {0} in {1}:'.format(self.parseVar(varValue), self.parseVar(varData)))
        else:
            if varValue is None:
                mainNode.writeCode('for {0} in range(len({1})):'.format(self.parseVar(varKey), self.parseVar(varData)))
            else:
                mainNode.writeCode('for {0},{1} in enumerate({2}):'.format(self.parseVar(varKey),self.parseVar(varValue), self.parseVar(varData)))

        self.parseNode(mainNode, subNode)

        if varIndex is not None:
            subNode.writeCode('{0}+=1'.format(varIndex),True,False)

    def _dict(self, node_attrs, mainNode: BaseNode, subNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        if not isinstance(attrs, dict):
            mainNode.writeCode('for {0}:'.format(node_attrs))
            self.parseNode(mainNode, subNode)
            return

        varData = attrs['name']
        varKey = attrs.get('key', None)
        varValue = attrs.get('value', None)
        varIndex = attrs.get('index', None)

        if varIndex is not None:
            mainNode.writeCode('{0} = 0'.format(varIndex), True, False)

        if varValue is None:
            mainNode.writeCode('for {0} in {1}:'.format(self.parseVar(varKey), self.parseVar(varData)))
        else:
            mainNode.writeCode(
                'for {0},{1} in {2}.items():'.format(self.parseVar(varKey), self.parseVar(varValue),
                                                        self.parseVar(varData)))
        self.parseNode(mainNode, subNode)

        if varIndex is not None:
            subNode.writeCode('{0}+=1'.format(varIndex),True,False)


    def _if(self, node_attrs, mainNode: BaseNode, subNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        if not isinstance(attrs, dict):
            condition = node_attrs
        else:
            condition = attrs['condition']

        mainNode.writeCode('if {0}:'.format(condition))


        # 继续解析node body
        self.parseNode(mainNode, subNode)

    def _elif(self, node_attrs, mainNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        if not isinstance(attrs, dict):
            condition = node_attrs
        else:
            condition = attrs['condition']

        mainNode.writeCode('elif {0}:'.format(condition),False)



    def _else(self, node_attrs, mainNode: BaseNode):

        mainNode.writeCode('else:',False)


    def _python(self,node_attrs, mainNode: BaseNode, subNode: BaseNode):
        linebody = subNode.getLineBody()

        # 查找空格最少的行
        mincount = 100000;
        for line in linebody:
            lcount = ViewUtil.countlstrip(line)

            if lcount is not False and lcount < mincount:
                mincount = lcount;

        for line in linebody:
            mainNode.writeCode(line[mincount:])

    def _comment(self,node_attrs, mainNode: BaseNode, subNode: BaseNode):

        pass

    def _pass(self,node_attrs, mainNode: BaseNode, subNode: BaseNode):

        mainNode.writeStr(subNode.getBody())

    def _set(self, node_attrs, mainNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        for name,value in attrs.items():
            mainNode.writeCode('{0} = {1}'.format(name,self.parseVar(value)))

    def _raw(self, node_attrs, mainNode: BaseNode):

        code = str(node_attrs).lstrip()
        mainNode.writeCode(code)


    def _empty(self, node_attrs, mainNode: BaseNode, subNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        if not isinstance(attrs, dict):
            condition = node_attrs
        else:
            condition = self.parseVar(attrs['name'])

        if_attrs = {
            "condition":'not {0}'.format(condition)
        }

        self._if(if_attrs,mainNode,subNode)

    def _notempty(self, node_attrs, mainNode: BaseNode, subNode: BaseNode):

        attrs = self.parseAttr(node_attrs)
        if not isinstance(attrs, dict):
            condition = node_attrs
        else:
            condition = self.parseVar(attrs['name'])

        if_attrs = {
            "condition": condition
        }

        self._if(if_attrs,mainNode,subNode)